<?php

namespace Drupal\gridstack;

use Drupal\Component\Utility\NestedArray;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Render\Element;
use Drupal\blazy\Blazy;
use Drupal\blazy\BlazyManagerBase;
use Drupal\gridstack\Entity\GridStack;

/**
 * Implements GridStackManagerInterface.
 */
class GridStackManager extends BlazyManagerBase implements GridStackManagerInterface {

  /**
   * Static cache for the skin definition.
   *
   * @var array
   */
  protected $skinDefinition;

  /**
   * Static cache for the skin options.
   *
   * @var array
   */
  protected $skinOptions;

  /**
   * Static cache for the optionset options.
   *
   * @var array
   */
  protected $optionsetOptions;

  /**
   * If should ungridstack.
   *
   * @var bool
   */
  protected $unGridStack = FALSE;

  /**
   * Returns defined skins as registered via hook_gridstack_skins_info().
   */
  public function getSkins() {
    if (!isset($this->skinDefinition)) {
      $this->skinDefinition = $this->buildSkins('gridstack', '\Drupal\gridstack\GridStackSkin');
    }

    return $this->skinDefinition;
  }

  /**
   * Returns available skins for select options.
   */
  public function getSkinOptions() {
    if (!isset($this->skinOptions)) {
      $skins = [];
      foreach ($this->getSkins() as $skin => $properties) {
        $skins[$skin] = isset($properties['name']) ? strip_tags($properties['name']) : $skin;
      }
      $this->skinOptions = $skins;
    }

    return $this->skinOptions;
  }

  /**
   * Returns available options for select options.
   */
  public function getOptionsetsByGroupOptions($group = '') {
    if (!isset($this->optionsetOptions)) {
      $optionsets = [];
      foreach ($this->entityLoadMultiple('gridstack') as $key => $entity) {
        // Exludes Boostrap/ Foundation grids which only work for DS, Panels.
        if ($group && $group == 'js' && $entity->getOption('use_framework')) {
          continue;
        }
        $optionsets[$key] = strip_tags($entity->label());
      }
      $this->optionsetOptions = $optionsets;
    }

    return $this->optionsetOptions;
  }

  /**
   * {@inheritdoc}
   */
  public function attach($attach = []) {
    $attach += ['skin' => FALSE];

    $load = parent::attach($attach);

    // Only load libraries if not destroyed.
    if (!$this->unGridStack) {
      // Only load GridStack JS if not using Bootstrap, nor Foundation.
      if (!empty($attach['use_js'])) {
        $customized = $this->configLoad('customized', 'gridstack.settings');
        $load['library'][] = empty($customized) ? 'gridstack/load' : 'gridstack/customized';

        if (!empty($attach['width']) && $attach['width'] < 12) {
          $load['library'][] = 'gridstack/gridstack.' . $attach['width'];
        }

        // Breakpoints: xs sm md lg requires separate CSS files.
        if (!empty($attach['breakpoints'])) {
          foreach ($attach['breakpoints'] as $breakpoint) {
            if (!empty($breakpoint['column']) && $breakpoint['column'] < 12) {
              $load['library'][] = 'gridstack/gridstack.' . $breakpoint['column'];
            }
          }
        }

        $load['drupalSettings']['gridstack'] = GridStack::load('default')->getOptions('settings');
      }
    }

    // The CSS framework grid library for admin pages such as layout builder.
    if (!empty($attach['library'])) {
      $load['library'][] = $attach['library'];
    }

    // Skins may be available for JS, or CSS layouts, or even ungridstack.
    if ($skin = $attach['skin']) {
      $skins = $this->getSkins();
      $provider = isset($skins[$skin]['provider']) ? $skins[$skin]['provider'] : 'gridstack';
      $load['library'][] = 'gridstack/' . $provider . '.' . $skin;
    }

    $this->moduleHandler->alter('gridstack_attach', $load, $attach);
    return $load;
  }

  /**
   * {@inheritdoc}
   */
  public function boxAttributes(array &$settings, $optionset, $current = 'grids') {
    // Allows extenders to provide own grid attributes.
    if ($this->unGridStack) {
      return [];
    }

    return $settings['use_js'] ? $this->jsBoxAttributes($settings, $optionset, $current) : $this->cssBoxAttributes($settings, $optionset, $current);
  }

  /**
   * Provides dynamic GridStack JS grid attributes.
   */
  public function jsBoxAttributes(array &$settings, $optionset, $current = 'grids') {
    $attributes  = [];
    $nameshort   = $settings['nameshort'];
    $breakpoint  = isset($settings['breakpoint']) ? $settings['breakpoint'] : 'lg';
    $breakpoints = isset($settings['breakpoints']) ? $settings['breakpoints'] : [];
    $id          = isset($settings['delta']) ? $settings['delta'] : 0;
    $nid         = isset($settings['nested_delta']) ? $settings['nested_delta'] : NULL;
    $end_grids   = $optionset->getEndBreakpointGrids($current);
    $grids       = isset($breakpoints[$breakpoint]) && isset($breakpoints[$breakpoint][$current]) ? $breakpoints[$breakpoint][$current] : $end_grids;

    // Nested grids.
    if (isset($settings['nested_delta']) && $current == 'nested') {
      foreach (['x', 'y', 'width', 'height'] as $key) {
        if (!isset($grids[$id][$nid])) {
          continue;
        }

        $attributes['data-' . $nameshort . '-' . $key] = isset($grids[$id][$nid][$key]) ? $grids[$id][$nid][$key] : 0;
      }
    }
    else {
      // The root element grids.
      foreach (['x', 'y', 'width', 'height'] as $key) {
        $attributes['data-' . $nameshort . '-' . $key] = isset($grids[$id][$key]) ? $grids[$id][$key] : 0;
      }
    }

    return $attributes;
  }

  /**
   * Provides static Bootstrap/ Foundation CSS grid attributes.
   */
  public function cssBoxAttributes(array &$settings, $optionset, $current = 'grids') {
    $id         = isset($settings['delta']) ? $settings['delta'] : 0;
    $nid        = isset($settings['nested_delta']) ? $settings['nested_delta'] : NULL;
    $framework  = $settings['framework'];
    $regions    = isset($settings['regions']) ? $settings['regions'] : [];
    $rid        = $current == 'nested' ? 'gridstack_' . $id . '_' . $nid : 'gridstack_' . $id;
    $region     = empty($regions[$rid]) ? [] : $regions[$rid];
    $attributes = [];
    $points     = GridStackDefault::breakpoints();

    // These are configured via Layout Builder or Panels UI.
    if ($region) {
      if (isset($region['attributes']) && !empty($region['attributes'])) {
        $attributes = $optionset::parseAttributes($region['attributes']);
        unset($settings['regions'][$rid]['attributes']);
      }

      if (!empty($region['wrapper_classes'])) {
        $classes = explode(' ', $region['wrapper_classes']);
        foreach ($classes as $class) {
          $attributes['class'][] = trim($class);
        }
        unset($settings['regions'][$rid]['wrapper_classes']);
      }
    }

    // Bootstrap 4 uses flexbox with `col` class, and has `xl` breakpoint.
    // @todo: Make it generic for other frameworks, or at least alterable.
    if ($framework == 'bootstrap') {
      $attributes['class'][] = 'col';
    }
    elseif ($framework == 'foundation') {
      unset($points['xs'], $points['xl']);
    }

    $unique = $optionset::optimizeGridWidths($settings, $current);
    foreach ($points as $point => $label) {
      if (!isset($unique[$point])) {
        continue;
      }

      $prefix = $suffix = '';
      if (strpos($framework, 'bootstrap') !== FALSE) {
        // Specific to XS: Bootstrap 3: col-xs-*, Bootstrap 4: col-*.
        $prefix = 'col-' . $point . '-';
        if ($framework == 'bootstrap' && $point == 'xs') {
          $prefix = 'col-';
        }
      }
      elseif ($framework == 'foundation') {
        $prefix = $label . '-';
        $suffix = ' columns';
      }

      $attributes['class'][] = $prefix . $unique[$point] . $suffix;
    }

    return $attributes;
  }

  /**
   * {@inheritdoc}
   */
  public function adminAttributes(array &$box, array &$attributes, array $settings, array $regions = [], $rid = NULL) {
    $region = $regions[$rid];

    // Layout Builder integration.
    if (!empty($settings['_ipe_layouts']) && !empty($region['#attributes']) && is_array($region['#attributes'])) {
      $attributes = NestedArray::mergeDeep($attributes, $region['#attributes']);
      $attributes['class'][] = 'layout-builder--layout__region layout__region layout__region--' . $rid;

      if (isset($region['layout_builder_add_block'])) {
        $box[] = $region['layout_builder_add_block'];
      }
    }

    // Panels IPE integration.
    if (!empty($settings['_ipe_panels']) && !empty($region['#prefix'])) {
      $box['#prefix'] = $region['#prefix'];

      foreach (Element::children($box) as $bid) {
        if (isset($region[$bid]['#attributes']['data-block-id'])) {
          $box[$bid]['#attributes']['data-block-id'] = $region[$bid]['#attributes']['data-block-id'];
        }
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function buildItems(array $build, array $regions = []) {
    $optionset = $build['optionset'];
    $settings  = $build['settings'];
    $grids     = $optionset->getEndBreakpointGrids();
    $content   = [];

    foreach ($build['items'] as $delta => $item) {
      $box        = isset($item['box']) ? $item['box'] : [];
      $attributes = isset($item['attributes']) ? $item['attributes'] : [];
      $settings   = isset($item['settings']) ? array_merge($settings, $item['settings']) : $settings;
      $rid        = 'gridstack_' . $delta;
      $clone      = [
        'box'     => $box,
        'caption' => isset($item['caption']) ? $item['caption'] : [],
      ];

      $settings['delta'] = $delta;
      if ($grids) {

        // Skips if more than we can chew, otherwise broken grid anyway.
        // The grids is a fixed layout blueprint. The items is dynamic contents.
        if (!isset($grids[$delta])) {
          continue;
        }

        // Layout Builder or Panels IPE only output for granted users.
        if (!empty($settings['_access_ipe']) && $regions && isset($regions[$rid])) {
          $this->adminAttributes($clone['box'], $attributes, $settings, $regions, $rid);
        }

        // Node contains the main grids/boxes.
        // @todo: See if to check for $settings['root'] for $nested_grids.
        $nested_grids = $optionset->getNestedGridsByDelta($delta);
        $is_nested = array_filter($nested_grids);
        $root_settings = $settings;

        // Overrides wrapper via UI, if so configured.
        if (isset($settings['regions']) && !empty($settings['regions'][$rid]['wrapper'])) {
          $root_settings['wrapper'] = $settings['regions'][$rid]['wrapper'];
        }

        $main_attributes = $this->boxAttributes($settings, $optionset, 'grids');
        $clone['attributes'] = NestedArray::mergeDeep($attributes, $main_attributes);
        $clone['settings'] = $root_settings;

        // Nested grids with preserved indices even if empty so to layout.
        // Only Bootstrap/ Foundation has nested grids, not js-driven layouts.
        if ($is_nested && !empty($box) && isset($box[0]['box'])) {
          $this->buildNestedItems($clone, $delta, $item, $optionset, $nested_grids, $root_settings, $settings, $regions);
        }
      }

      $content[] = $clone;
      unset($clone);
    }

    return $content;
  }

  /**
   * Provides nested items if so configured.
   */
  public function buildNestedItems(array &$clone, $delta, $item, $optionset, $grids, $root_settings, $settings, $regions = []) {
    $nested = $item['box'];
    $settings['root'] = FALSE;
    $settings['nested'] = $root_settings['nested'] = TRUE;

    $clone['attributes']['class'][] = 'box--nester';

    // The nested elements.
    foreach (array_keys($grids) as $gid) {
      $rid = 'gridstack_' . $delta . '_' . $gid;
      $settings['nested_delta'] = $gid;
      $nested[$gid]['settings'] = isset($nested[$gid]['settings']) ? array_merge($settings, $nested[$gid]['settings']) : $settings;

      if (isset($settings['regions'][$rid]) && !empty($settings['regions'][$rid]['wrapper'])) {
        $nested[$gid]['settings']['wrapper'] = $settings['regions'][$rid]['wrapper'];
      }

      $nested[$gid]['settings']['nested_id'] = ($delta + 1) . '-' . ($gid + 1);
      $attributes = isset($nested[$gid]['attributes']) ? $nested[$gid]['attributes'] : [];

      // Layout Builder or Panels IPE integration only output for granted users.
      if ($settings['_access_ipe'] && $regions && isset($regions[$rid])) {
        $this->adminAttributes($nested[$gid]['box'], $attributes, $settings, $regions, $rid);
      }

      $nested[$gid]['attributes'] = NestedArray::mergeDeep($attributes, $this->boxAttributes($settings, $optionset, 'nested'));
      $nested[$gid]['attributes']['class'][] = 'box--nested';
    }

    // Adds prefix for admin actions, etc.
    $box = isset($item['prefix']) ? ['prefix' => $item['prefix']] : [];

    // Provides nested gridstack, gridstack within gridstack, if so configured.
    $box['content'] = [
      '#theme'      => 'gridstack',
      '#items'      => $nested,
      '#optionset'  => $optionset,
      '#settings'   => $settings,
      '#attributes' => isset($item['wrapper_attributes']) ? $item['wrapper_attributes'] : [],
    ];

    $clone['box'] = $box;
    $clone['settings'] = $root_settings;
  }

  /**
   * Provides multi-breakpoint image styles.
   *
   * Overrides fallback breakpoint image_style with grid image_style.
   * This tells theme_blazy() to respect different image style per item.
   */
  public function buildImageStyleMultiple(array &$settings, $delta = 0) {
    foreach ($settings['breakpoints'] as $key => &$breakpoint) {
      if (isset($breakpoint['image_style']) && !empty($breakpoint['grids'][$delta]) && !empty($breakpoint['grids'][$delta]['image_style'])) {
        $breakpoint['image_style'] = $breakpoint['grids'][$delta]['image_style'];
      }

      // Overrides image style to use a defined image style per grid item.
      // This allows each individual box to have different image style.
      if ($key == 'lg' && !empty($breakpoint['grids'][$delta]['image_style'])) {
        $settings['_dimensions'] = FALSE;
        $settings['image_style'] = $breakpoint['grids'][$delta]['image_style'];
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function build(array $build = []) {
    if (empty($build['items'])) {
      return [];
    }

    foreach (['attached', 'grids', 'items', 'layout', 'optionset', 'settings'] as $key) {
      $build[$key] = isset($build[$key]) ? $build[$key] : [];
    }

    $settings = &$build['settings'];
    $id = empty($settings['id']) ? 'gridstack-' . $settings['optionset'] : $settings['id'];
    $settings['id'] = Blazy::getHtmlId('gridstack', $id);

    $gridstack = [
      '#theme'      => 'gridstack',
      '#items'      => [],
      '#build'      => $build,
      '#pre_render' => [[$this, 'preRenderGridStack']],
    ];

    // Prepare wasted empty children to support Panels, and Layout Builder.
    if ($layout = $build['layout']) {
      foreach (array_keys($layout->getRegions()) as $rid) {
        $gridstack[$rid] = [];
      }
    }

    // Provides cache under production mode, or if so configured.
    $gridstack['#cache'] = $this->getCacheMetadata($build);
    $this->moduleHandler->alter('gridstack_build', $gridstack, $settings);

    return $gridstack;
  }

  /**
   * Return the cache metadata.
   */
  private function getCacheMetadata(array $build = []) {
    $settings          = $build['settings'];
    $max_age           = $this->configLoad('cache.page.max_age', 'system.performance');
    $max_age           = empty($settings['cache']) ? $max_age : $settings['cache'];
    $id                = $settings['id'];
    $suffixes[]        = count($build['items']);
    $suffixes[]        = count(array_filter($settings));
    $suffixes[]        = $max_age;
    $cache['tags']     = Cache::buildTags('gridstack:' . $id, $suffixes, '.');
    $cache['contexts'] = ['languages'];
    $cache['max-age']  = $max_age;
    $cache['keys']     = isset($settings['cache_metadata']['keys']) ? $settings['cache_metadata']['keys'] : [$id];

    if (!empty($settings['cache_tags'])) {
      $cache['tags'] = Cache::mergeTags($cache['tags'], $settings['cache_tags']);
    }

    return $cache;
  }

  /**
   * Return the wrapper attributes.
   */
  public function prepareAttributes(array $build, $optionset) {
    $settings = $build['settings'];
    $attributes = empty($settings['attributes']) ? [] : $optionset::parseAttributes($settings['attributes']);
    $blazy = empty($settings['blazy_data']) ? '' : Json::encode($settings['blazy_data']);

    $attributes['class'][] = 'blazy';
    $attributes['data-blazy'] = $blazy;

    if (!empty($settings['debug'])) {
      $attributes['class'][] = 'gridstack--debug';
    }

    if ($settings['media_switch']) {
      $switch = str_replace('_', '-', $settings['media_switch']);
      $attributes['data-' . $switch . '-gallery'] = TRUE;
    }

    // Only add attributes if using a gridstack, save for few.
    $attributes['id'] = $settings['id'];
    if (!$this->unGridStack) {
      // Adds wrapper classes for static grid Bootstrap/ Foundation.
      if ($settings['use_framework']) {
        if (!empty($settings['wrapper_classes']) && is_string($settings['wrapper_classes'])) {
          $classes = explode(' ', $settings['wrapper_classes']);
          foreach ($classes as $class) {
            $attributes['class'][] = trim($class);
          }
        }
      }
      elseif ($settings['use_js']) {
        // Adds attributes for js-driven layouts.
        // Gets options.breakpoints.sm.[width, column, image_style, grids], etc.
        $exclude_image_style = empty($settings['_admin']);
        $columns = $optionset->getJson('breakpoints');
        $js = $optionset->getJson('settings');
        if ($responsives = array_filter($optionset->getBreakpoints())) {
          foreach ($optionset::getConstantBreakpoints() as $breakpoint) {
            $responsive_grids = $optionset->getJsonSummaryBreakpoints($breakpoint, $exclude_image_style);
            $has_width = isset($responsives[$breakpoint]['width']) && $responsives[$breakpoint]['width'] > -1;
            if ($has_width && $responsive_grids) {
              $attributes['data-' . $breakpoint . '-width'] = $responsives[$breakpoint]['width'];
              $attributes['data-' . $breakpoint . '-grids'] = $responsive_grids;
            }
          }
        }

        // Add the required configuration as JSON object.
        $attributes['data-breakpoints'] = strpos($columns, '{"":12}') !== FALSE ? '' : $columns;
        $attributes['data-config']      = $js;

        // Breakpoint related data-attributes helpers.
        if (!empty($settings['minWidth'])) {
          $attributes['data-min-width'] = $settings['minWidth'];
        }
      }
    }

    return $attributes;
  }

  /**
   * Load the optionset with a fallback.
   */
  public function load($id) {
    $optionset = GridStack::load($id);
    if (empty($optionset)) {
      $optionset = GridStack::load('default');
    }
    return $optionset;
  }

  /**
   * Build the HTML settings.
   */
  public function prepareSettings(&$settings, $optionset) {
    $defaults = array_merge(GridStackDefault::htmlSettings(), $optionset->getOptions('settings'));
    $settings += $defaults;

    // Use static grid framework if so configured.
    $settings['framework']     = $this->configLoad('framework', 'gridstack.settings');
    $settings['debug']         = $this->configLoad('debug', 'gridstack.settings');
    $settings['use_framework'] = !empty($settings['framework']) && $optionset->getOption('use_framework');

    // Disable background and JS if using a CSS framework.
    if ($settings['use_framework']) {
      $settings['background'] = $settings['use_js'] = FALSE;
      // Admin UI always uses gridstack JS to build layouts regardless of this.
      $settings['use_framework'] = empty($settings['_admin']);
    }
    else {
      $settings['use_js'] = !empty($settings['root']) || !empty($settings['_admin']);
    }

    if (empty($settings['breakpoints'])) {
      $optionset->gridsJsonToArray($settings);
    }

    // Do not leak container attributes as already processed above.
    $settings['attributes'] = $settings['wrapper_classes'] = '';
  }

  /**
   * {@inheritdoc}
   */
  public function preRenderGridStack($element) {
    $build = $element['#build'];
    unset($element['#build']);

    if (empty($build['items'])) {
      return [];
    }

    // Build gridstack elements.
    $settings = &$build['settings'];
    $optionset = $build['optionset'] ?: GridStack::load($settings['optionset']);

    // Supports Blazy multi-breakpoint images if provided.
    if (!empty($settings['check_blazy']) && !empty($build['items'][0])) {
      $this->isBlazy($settings, $build['items'][0]);
    }

    $this->prepareSettings($settings, $optionset);

    // Use static grid framework if so configured.
    $settings['ungridstack']  = $this->unGridStack;
    $settings['_ipe_layouts'] = !empty($settings['_ipe_layouts']) && isset($element['#attributes']) && isset($element['#attributes']['data-layout-delta']);
    $settings['_ipe_panels']  = !empty($settings['_ipe_panels']) && isset($element['#prefix']) && strpos($element['#prefix'], 'panels-ipe-content') !== FALSE;
    $settings['_access_ipe']  = $is_admin = $settings['_ipe_layouts'] || $settings['_ipe_panels'];
    $settings['library']      = $is_admin ? $this->configLoad('library', 'gridstack.settings') : '';

    // Adds attachments.
    $attachments = $this->attach($settings);

    // Adds wrapper attributes fo the root element only.
    if ($settings['root']) {
      $element['#attributes'] = $this->prepareAttributes($build, $optionset);
    }

    // Provides regions for Layout Builder and Panels IPE integration.
    $regions = [];
    $children = Element::children($element);
    if ($is_admin) {
      foreach ($children as $child) {
        if ($child == 'items') {
          continue;
        }
        $regions[$child] = $element[$child];
      }

      $build['attached']['library'][] = 'gridstack/admin_base';
    }

    $element['#optionset'] = $build['optionset'] = $optionset;
    $element['#settings']  = $build['settings'] = $settings;
    $element['#attached']  = empty($build['attached']) ? $attachments : NestedArray::mergeDeep($build['attached'], $attachments);
    $element['#items']     = $this->buildItems($build, $regions);

    // Panels IPE, Layout Builder are happy, safe to free up wasted children.
    if ($children) {
      foreach ($children as $child) {
        unset($element[$child]);
      }
    }

    return $element;
  }

}
